from ._accessformatter import AccessFormatter
from ._colourizedformatter import ColourizedFormatter
from ._defaultformatter import DefaultFormatter
from ._jsonformatter import JSONFormatter


__all__: list[str] = [
    'AccessFormatter',
    'ColourizedFormatter',
    'DefaultFormatter',
    'JSONFormatter',
]


from typing import Any
from typing import Callable
from typing import Generic
from typing import TypeVar


C = TypeVar('C')
T = TypeVar('T')


class _classproperty(Generic[T]):
    func: Callable[..., T]

    def __init__(self, func: Any):
        self.func = func # type: ignore

    def __get__(self, _: Any, cls: Any) -> T:
        return self.func(cls)


def classproperty(
    func: Callable[..., T]
) -> _classproperty[T]:
    return _classproperty(func)


class MissingDependency:

    def __init__(self, message: str):
        self.message = message

    def __getattr__(self, attname: str) -> Any:
        raise ImportError(self.message)


from ._conflict import Conflict
from ._doesnotexist import DoesNotExist
from ._fatalexception import FatalException
from ._undecryptable import Undecryptable


__all__: list[str] = [
    'Conflict',
    'ExceptionRaiser',
    'DoesNotExist',
    'FatalException',
    'Undecryptable',
]


class ExceptionRaiser:
    """Mixin class that provides an interface to raise standard exceptions
    defined by the :mod:`libcanonical.types.exceptions` package.
    """
    __module__: str = 'libcanonical.types'
    Conflict        = Conflict
    DoesNotExist    = DoesNotExist
    Undecryptable   = Undecryptable


from typing import Any
from typing import cast
from typing import Self

import langcodes
from langcodes import Language
from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema
from pydantic_core import core_schema


class BCP47(Language):
    __module__: str = 'canonical.ext.i18n.types'

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(),
            python_schema=core_schema.union_schema([
                core_schema.no_info_plain_validator_function(cls),
                core_schema.chain_schema([
                    core_schema.is_instance_schema(str),
                ])
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema())

    @classmethod
    def fromstring(cls, value: str) -> Self:
        value = langcodes.standardize_tag(value, macro=True)
        return cast(Self, cls.get(value))


import ipaddress
import os
import pathlib
from typing import Literal

import pydantic
from jinja2 import Template
from jinja2 import StrictUndefined


class EnvironmentVariables(pydantic.BaseModel):
    debug: bool = pydantic.Field(
        alias='DEBUG',
        default=False
    )

    deployment_env: Literal['development', 'testing', 'staging', 'production'] = pydantic.Field(
        alias='DEPLOYMENT_ENV',
        default='production'
    )

    name: str = pydantic.Field(
        default=...
    )

    loglevel: str = pydantic.Field(
        alias='LOGLEVEL',
        default='INFO'
    )

    etcdir: pathlib.Path = pydantic.Field(
        alias='ETCDIR',
        default=...
    )

    vardir: pathlib.Path = pydantic.Field(
        alias='VARDIR',
        default=...
    )

    #: The bind address of the primary HTTP interface.
    http_bind_address: ipaddress.IPv4Address = pydantic.Field(
        alias='HTTP_HOST',
        default=ipaddress.IPv4Address('127.0.0.1')
    )

    #: The port of the primary HTTP interface.
    http_bind_port: int = pydantic.Field(
        alias='HTTP_PORT',
        default=8000
    )

    @pydantic.model_validator(mode='before')
    def validate_params(cls, params: dict[str, str]):
        name = params.get('name')
        if not isinstance(name, str):
            raise pydantic.ValidationError()
        if not params.get('VARDIR'):
            params['ETCDIR'] = f'/etc/{name}'
        if not params.get('VARDIR'):
            params['VARDIR'] = f'/var/lib/{name}'
        return params

    @classmethod
    def parse(cls, app_name: str):
        return cls.model_validate({**os.environ, 'name': app_name})

    def parse_config_file(self, filename: str) -> str:
        """Parses a configuration file relative to :attr:`etcdir`."""
        t = Template(
            source=open(self.etcdir.joinpath(filename)).read(),
            variable_start_string='${',
            variable_end_string='}',
            undefined=StrictUndefined
        )
        return t.render(
            env={
                **os.environ,
                **self.model_dump(
                    mode='json',
                    by_alias=True,
                    exclude_none=True
                )
            }
        )


defaults = EnvironmentVariables.parse('.canonical')


import sys

from ._colourizedformatter import ColourizedFormatter


class DefaultFormatter(ColourizedFormatter):
    __module__: str = 'libcanonical.utils.logging'

    def should_use_colors(self) -> bool:
        return sys.stderr.isatty()  # pragma: no cover


from ._environmentbasemodel import EnvironmentBaseModel
from ._pollingexternalstate import PollingExternalState
from ._regexclassifier import RegexClassifier
from ._statelogger import StateLogger
from ._taggable import Taggable


__all__: list[str] = [
    'EnvironmentBaseModel',
    'PollingExternalState',
    'RegexClassifier',
    'StateLogger',
    'Taggable'
]


import json
from typing import Any
from typing import ClassVar
from typing import TypeVar

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler

from libcanonical.utils.encoding import b64decode
from libcanonical.utils.encoding import b64encode
from libcanonical.utils.encoding import b64encode_int
from libcanonical.utils.encoding import bytes_to_number


__all__: list[str] = [
    'Base64'
]

T = TypeVar('T', bound='Base64')


class Base64(bytes):
    __module__: str = 'libcanonical.types'
    description: ClassVar[str | None] = None
    type_name: ClassVar[str | None] = None
    max_length: ClassVar[int | None] = None
    min_length: ClassVar[int | None] = None

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.no_info_plain_validator_function(cls.fromb64),
            python_schema=core_schema.union_schema([
                core_schema.chain_schema([
                    core_schema.is_instance_schema(cls),
                    core_schema.bytes_schema(
                        max_length=cls.max_length,
                        min_length=cls.min_length
                    ),
                    core_schema.no_info_plain_validator_function(cls)
                ]),
                core_schema.chain_schema([
                    core_schema.is_instance_schema(str),
                    core_schema.no_info_plain_validator_function(cls.fromb64),
                    core_schema.bytes_schema(
                        max_length=cls.max_length,
                        min_length=cls.min_length
                    ),
                    core_schema.no_info_plain_validator_function(cls)
                ]),
                core_schema.chain_schema([
                    core_schema.is_instance_schema(int),
                    core_schema.no_info_plain_validator_function(cls.fromint),
                    core_schema.bytes_schema(
                        max_length=cls.max_length,
                        min_length=cls.min_length
                    ),
                    core_schema.no_info_plain_validator_function(cls)
                ]),
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        schema = handler(core_schema.str_schema())
        if cls.description:
            schema['description'] = cls.description
        if cls.type_name:
            schema['title'] = cls.type_name
        return schema

    @classmethod
    def b64decode(cls, value: bytes | str):
        return b64decode(value)

    @classmethod
    def b64encode(cls, value: bytes | str) -> str:
        return  b64encode(value, encoder=str)

    @classmethod
    def fromint(cls, value: int):
        return cls(b64decode(b64encode_int(value)))

    @classmethod
    def fromb64(cls, value: bytes | str):
        return cls(b64decode(value))

    @classmethod
    def fromdict(cls, value: dict[str, Any]):
        return cls(str.encode(json.dumps(value), 'utf-8'))

    @classmethod
    def fromstring(cls, value: str):
        return cls(str.encode(value))

    @classmethod
    def validate(cls, instance: T) -> T:
        return instance

    def is_empty(self):
        return not bool(self)

    def urlencode(self):
        return b64encode(self)

    def __str__(self):
        return b64encode(self, encoder=str)

    def __int__(self):
        return bytes_to_number(self)

    def __repr__(self):
        return str(self)


import asyncio
import functools
import inspect
import logging
import threading
import time
from typing import Callable
from typing import Iterable
from typing import ParamSpec
from typing import TypeVar


P = ParamSpec('P')
T = TypeVar('T')


class _Retry:
    logger: logging.Logger = logging.getLogger('canonical')
    exception_classes: tuple[type[BaseException], ...]
    logging_message: str = "Caught retryable %(exception_class)s (thread: %(thread)s, attempts: %(attempts)s)"

    def __init__(
        self,
        types: Iterable[type[BaseException]],
        max_attempts: int | None = None,
        delay: int | float = 10,
        reason: str | None = None
    ):
        self.delay = delay
        self.exception_classes = tuple(types)
        self.max_attempts = max_attempts
        if reason:
            self.logging_message = reason

    def on_exception(self, exception: BaseException, attempts: int):
        if not isinstance(exception, self.exception_classes):
            return True
        if self.max_attempts and attempts > self.max_attempts:
            return True
        t = threading.current_thread()
        self.logger.warning(
            self.logging_message,
            extra={
                'attempts': attempts,
                'exception_class': f'{type(exception).__name__}',
                'thread_id': t.ident
            }
        )

    def call_sync(self, f: Callable[P, T]) -> Callable[P, T]:
        handles = self.exception_classes

        @functools.wraps(f)
        def d(*args: P.args, **kwargs: P.kwargs):
            attempts = 0
            while True:
                attempts += 1
                try:
                    result = f(*args, **kwargs)
                    break
                except KeyboardInterrupt:
                    raise
                except Exception as e:
                    if not isinstance(e, handles):
                        raise
                    if self.max_attempts and attempts > self.max_attempts:
                        raise e from e
                    t = threading.current_thread()
                    self.logger.warning(
                        self.logging_message,
                        extra={
                            'attempts': attempts,
                            'exception_class': f'{type(e).__name__}',
                            'thread_id': t.ident
                        }
                    )
                    time.sleep(self.delay)
            return result

        return d # type: ignore

    def __call__(
        self,
        f: Callable[P, T]
    ) -> Callable[P, T]:
        if not inspect.iscoroutinefunction(f):
            return self.call_sync(f)
        assert inspect.iscoroutinefunction(f)
        handles = self.exception_classes

        @functools.wraps(f)
        async def d(*args: P.args, **kwargs: P.kwargs):
            attempts = 0
            while True:
                attempts += 1
                try:
                    result = await f(*args, **kwargs)
                    break
                except KeyboardInterrupt:
                    raise
                except Exception as e:
                    if not isinstance(e, handles):
                        raise
                    if self.max_attempts and attempts > self.max_attempts:
                        raise e from e
                    t = threading.current_thread()
                    self.logger.warning(
                        self.logging_message,
                        extra={
                            'attempts': attempts,
                            'exception_class': f'{type(e).__name__}',
                            'thread_id': t.ident
                        }
                    )
                    await asyncio.sleep(self.delay)
            return result

        return d # type: ignore


retry = _Retry


from typing import Iterable

from .mediatype import get_best_match


class MediaTypeSelector:

    def __init__(self, allow: Iterable[str]):
        self.allow = set(allow)

    def select(self, header: str | None):
        if header is None:
            return None
        return get_best_match(header, self.allow)


from ._base import BaseCanonicalException


class Conflict(BaseCanonicalException):
    """This exception may be raised when a destructive operation is
    attempted on a resource, that consitutes a data integrity contstaint
    violation.
    """
    http_status_code: int = 409


import functools
import urllib.parse
from typing import Any

from .domainname import DomainName
from .stringtype import StringType


class HTTPResourceLocator(StringType):
    __module__: str = 'libcanonical.types'
    max_length = 2048
    protocols: set[str] = {'http', 'https'}
    description = (
        "An RFC HTTP resource locator, defined in RFC 3986 and RFC 9110, "
        "is a Uniform Resource Locator (URL) that specifies the location "
        "of a resource on the web, consisting of a scheme (http or https), "
        "an authority (domain or IP), an optional port, a path, an optional "
        "query, and an optional fragment (e.g., `https://example.com:8080/pa"
        "th?query=1#section`)."
    )

    @functools.cached_property
    def parts(self) -> urllib.parse.ParseResult:
        return urllib.parse.urlparse(str(self))

    @classmethod
    def validate(cls, v: str, _: Any = None):
        p = urllib.parse.urlparse(v)
        if p.scheme not in cls.protocols:
            raise ValueError(f"Not a valid URL: {v[:128]}")
        return cls(v)

    @property
    def query(self) -> dict[str, str | list[str]]:
        p =urllib.parse.urlparse(self)
        values: dict[str, list[str] | str] = {}
        for name, value in urllib.parse.parse_qs(p.query).items():
            if len(value) == 1:
                value = value[0]
            values[name] = value
        return values

    def is_subpath(self, value: str) -> bool:
        return self.startswith(value)

    def with_query(self, **kwargs: Any) -> 'HTTPResourceLocator':
        p: list[str] = list(urllib.parse.urlparse(self))
        q = dict(urllib.parse.parse_qs(p[4]))
        q.update(kwargs)
        p[4] = urllib.parse.urlencode(q, doseq=True)
        return HTTPResourceLocator(urllib.parse.urlunparse(p))
        


import logging


class TracebackInfoFilter(logging.Filter):
    """Clear or restore the exception on log records"""

    def __init__(self, clear: bool = True):
        self.clear = clear

    def filter(self, record: logging.LogRecord):
        if self.clear:
            record.exc_info = None
            record.exc_text = None
        return True


from ._baseencryptionresult import BaseEncryptionResult


class BytesEncryptionResult(BaseEncryptionResult):
    """The result of an encryption operation using the AES algorithm."""

    @property
    def aad(self) -> bytes:
        return b''

    @property
    def iv(self) -> None:
        return None

    @property
    def tag(self) -> None:
        return None

    def __bytes__(self) -> bytes:
        return self.ct